selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))

# Make
SHELL := /bin/bash
subMake = $(MAKE) $(@F) --directory=$(@D)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# Terminal
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
CP = cp -p
DIFF = diff --unified=2

tablesSort := plots organisms stems
vegxMaps := $(wildcard $(tablesSort:%=maps/VegX.%.csv))
vegxMaps += $(filter-out $(vegxMaps),$(wildcard maps/VegX.*.csv))
tables := $(vegxMaps:maps/VegX.%.csv=%)
vegbienMaps := $(subst VegX.,VegBIEN.,$(vegxMaps))

+maps = $(filter maps/%,$(+_))
+in = $(filter-out maps/%,$(+_))

root := $(selfDir_uZPPqC)..
psqlAsBien := $(root)/bin/psql_vegbien

# Config
test_n ?= 2

#####

all: _always maps ;

.SUFFIXES:

_always:
.PHONY: _always

clean: _always
	$(RM) $(all)

remake: _always clean all ;

%.out: %.make _always
	./$* >$@
.PRECIOUS: %.out

$(root)/%: _always
	+$(subMake)

#####

reinstall: _always uninstall install ;

#####

maps: $(vegbienMaps) _always ;

all += $(vegbienMaps)

maps/VegBIEN.%.csv: maps/VegX.%.csv $(root)/mappings/VegX-VegBIEN.%.csv
	$(root)/bin/join_sort <$+ >$@
.PRECIOUS: maps/VegBIEN.%.csv

#####

import: _always $(addprefix import-,$(tables)) verify ;

log = $*$(if $(n),.n=$(n),).log

import-%: maps/VegBIEN.%.csv _always
	-$(import_cmd)
# ignore import errors, which are often benign (e.g. invalid date format)
import_cmd = (set -x; "time" env commit=1 $(import) $<) \
$(if $(n),,>>$(log)) 2>&1$(if $(n),|tee -a $(log))

logs := $(wildcard *.log)

rm_logs: _always
	$(RM) $(logs)

#####

ifneq ($(wildcard verify.ref*),)

verify: verify.ref verify.out _always
	$(DIFF) $(+_)

all += verify.out

%.out: %.sql _always
	$(out_cmd)
.PRECIOUS: %.out
out_cmd = $(psqlAsBien) --no-align --field-separator='	' --pset=footer=off \
--pset=null=NULL <$< >$@

else
verify: _always ;
endif

#####

ifneq ($(wildcard test/),)

tests := test/VegX.xml test/VegBIEN.xml test/VegBIEN.2-step.xml test/import.out

test: _always $(tests) ;

.PRECIOUS: $(tests)

all += $(tests)

testRef = $(1:.2-step.xml=.xml).ref

define test
env test=1 n=$(test_n) $(1) $(if $(+in),<$(+in)) $(+maps) >$@
@(set -x; $(DIFF) $(call testRef,$@) $@) 2>&1 || { e=$$?;\
echo $(emph)"To accept new test output:"$(endEmph);\
echo "$(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile";\
exit $$e;}
endef

test/VegX.xml: $(vegxMaps) _always
	$(call test,$(map))

test/VegBIEN.xml: $(vegbienMaps) _always
	$(call test,$(map))

test/VegBIEN.2-step.xml: test/VegX.xml\
$(root)/mappings/VegX-VegBIEN.organisms.csv _always
	$(call test,$(root)/map)

test/import.out: $(vegbienMaps) _always
	$(call test,$(import))

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	$(CP) $* $(call testRef,$*)

else
test: _always ;
endif

#####

# Each input type needs var $(mapEnv) and targets install, uninstall

dbFile := $(firstword $(wildcard src/db.*.sql))

ifneq ($(dbFile),)

dbEngineExt := $(subst .,,$(suffix $(basename $(notdir $(dbFile)))))
db := $(notdir $(realpath .))

%.ref: %.ref.sql
	$(inputDbAsBien) $(db) <$< >$@
.PRECIOUS: %.ref

####

install: _always db ;

uninstall: _always rm_db ;

####

# Each DB engine needs vars $(dbEngine), $(dbAsBien) and targets db, rm_db

ifeq ($(dbEngineExt),my)

dbEngine := MySQL

bienPassword := $(shell cat $(root)/config/bien_password)
mysqlAs = mysql --user=$(1) --password='$(bienPassword)'
mysqlAsRoot := $(call mysqlAs,root)
dbAsBien := $(call mysqlAs,bien)

db: $(dbFile) _always
	echo "CREATE DATABASE $(db) DEFAULT CHARACTER SET latin1;"|$(mysqlAsRoot)
	-$(mysqlAsRoot) --database=$(db) <$<
	echo "GRANT SELECT ON $(db).* TO 'bien'@'localhost';"|$(mysqlAsRoot)
# ignore errors in db import so that GRANT will still be run

rm_db: _always
	-echo "REVOKE ALL ON $(db).* FROM 'bien'@'localhost';"|$(mysqlAsRoot)
	echo "DROP DATABASE IF EXISTS $(db);"|$(mysqlAsRoot)
# ignore errors if grant not defined

###

else
$(error The DB filename $(dbFile) must have the form db.{my|pg}.sql)
endif

# Must come after dbEngine is set
mapEnv := in_engine=$(dbEngine) in_database=$(db)

####

else
mapEnv :=
install: _always ;
uninstall: _always ;
endif

#####

# Must come after mapEnv is set
map := $(if $(mapEnv),env $(mapEnv) ,)$(root)/map
import := env out_database=vegbien $(map)
